// Copyright (c) [2024] SUSE LLC
//
// All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, contact SUSE LLC.
//
// To contact SUSE LLC about this file by physical or electronic mail, you may
// find current contact information at www.suse.com.

use std::{collections::HashMap, pin::Pin, task::Poll};

use agama_lib::{
    dbus::get_optional_property,
    error::ServiceError,
    jobs::{client::JobsClient, Job},
    property_from_dbus,
};
use axum::{extract::State, routing::get, Json, Router};
use futures_util::{ready, Stream};
use pin_project::pin_project;
use tokio::sync::mpsc::unbounded_channel;
use tokio_stream::{wrappers::UnboundedReceiverStream, StreamExt};
use zbus::zvariant::{ObjectPath, OwnedObjectPath, OwnedValue};

use crate::{
    dbus::{DBusObjectChange, DBusObjectChangesStream, ObjectsCache},
    error::Error,
    web::Event,
};

/// Builds a router for the jobs objects.
pub async fn jobs_service<T>(
    dbus: &zbus::Connection,
    destination: &'static str,
    path: &'static str,
) -> Result<Router<T>, ServiceError> {
    let client = JobsClient::new(dbus.clone(), destination, path).await?;
    let state = JobsState { client };
    Ok(Router::new().route("/jobs", get(jobs)).with_state(state))
}

#[derive(Clone)]
struct JobsState<'a> {
    client: JobsClient<'a>,
}

async fn jobs(State(state): State<JobsState<'_>>) -> Result<Json<Vec<Job>>, Error> {
    let jobs = state
        .client
        .jobs()
        .await?
        .into_iter()
        .map(|(_path, job)| job)
        .collect();
    Ok(Json(jobs))
}

/// Returns the stream of jobs-related events.
///
/// The stream combines the following events:
///
/// * Changes on the DASD devices collection.
///
/// * `dbus`: D-Bus connection to use.
pub async fn jobs_stream(
    dbus: zbus::Connection,
    destination: &'static str,
    manager: &'static str,
    namespace: &'static str,
) -> Result<Pin<Box<dyn Stream<Item = Event> + Send>>, Error> {
    let stream = JobsStream::new(&dbus, destination, manager, namespace).await?;
    Ok(Box::pin(stream))
}

#[pin_project]
pub struct JobsStream {
    dbus: zbus::Connection,
    cache: ObjectsCache<Job>,
    #[pin]
    inner: UnboundedReceiverStream<DBusObjectChange>,
}

#[derive(Debug, thiserror::Error)]
enum JobsStreamError {
    #[error("Service error: {0}")]
    Service(#[from] ServiceError),
    #[error("Unknown job: {0}")]
    UnknownJob(OwnedObjectPath),
}

impl JobsStream {
    pub async fn new(
        dbus: &zbus::Connection,
        destination: &'static str,
        manager: &'static str,
        namespace: &'static str,
    ) -> Result<Self, ServiceError> {
        let (tx, rx) = unbounded_channel();
        let mut stream = DBusObjectChangesStream::new(
            dbus,
            &ObjectPath::from_static_str(manager)?,
            &ObjectPath::from_static_str(namespace)?,
            "org.opensuse.Agama.Storage1.Job",
        )
        .await?;

        tokio::spawn(async move {
            while let Some(change) = stream.next().await {
                let _ = tx.send(change);
            }
        });
        let rx = UnboundedReceiverStream::new(rx);

        let mut cache: ObjectsCache<Job> = Default::default();
        let client = JobsClient::new(dbus.clone(), destination, manager).await?;
        for (path, job) in client.jobs().await? {
            cache.add(path, job);
        }

        Ok(Self {
            dbus: dbus.clone(),
            cache,
            inner: rx,
        })
    }

    fn update_job<'a>(
        cache: &'a mut ObjectsCache<Job>,
        path: &OwnedObjectPath,
        values: &HashMap<String, OwnedValue>,
    ) -> Result<&'a Job, ServiceError> {
        let job = cache.find_or_create(path);
        job.id = path.to_string();
        property_from_dbus!(job, running, "Running", values, bool);
        property_from_dbus!(job, exit_code, "ExitCode", values, u32);
        Ok(job)
    }

    fn remove_job(
        cache: &mut ObjectsCache<Job>,
        path: &OwnedObjectPath,
    ) -> Result<Job, JobsStreamError> {
        cache
            .remove(path)
            .ok_or_else(|| JobsStreamError::UnknownJob(path.clone()))
    }

    fn handle_change(
        cache: &mut ObjectsCache<Job>,
        change: &DBusObjectChange,
    ) -> Result<Event, JobsStreamError> {
        match change {
            DBusObjectChange::Added(path, values) => {
                let job = Self::update_job(cache, path, values)?;
                Ok(Event::JobAdded { job: job.clone() })
            }
            DBusObjectChange::Changed(path, updated) => {
                let job = Self::update_job(cache, path, updated)?;
                Ok(Event::JobChanged { job: job.clone() })
            }
            DBusObjectChange::Removed(path) => {
                let job = Self::remove_job(cache, path)?;
                Ok(Event::JobRemoved { job })
            }
        }
    }
}

impl Stream for JobsStream {
    type Item = Event;

    fn poll_next(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Option<Self::Item>> {
        let mut pinned = self.project();

        Poll::Ready(loop {
            let change = ready!(pinned.inner.as_mut().poll_next(cx));
            let next_value = match change {
                Some(change) => {
                    if let Ok(event) = Self::handle_change(pinned.cache, &change) {
                        Some(event)
                    } else {
                        log::warn!("Could not process change {:?}", &change);
                        None
                    }
                }
                None => break None,
            };
            if next_value.is_some() {
                break next_value;
            }
        })
    }
}
